// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_side_breakthrough.h"

#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>

using namespace rcsc;

/*-------------------------------------------------------------------*/
/*!

 */
bool
Bhv_SideBreakThrough::execute( PlayerAgent * agent )
{
	dlog.addText( Logger::TEAM,__FILE__": Bhv_SideBreakthrough" );

	const WorldModel & wm = agent->world();
	const PlayerPtrCont & opps = wm.opponentsFromSelf();
	const PlayerObject * nearest_opp= ( opps.empty()? static_cast< PlayerObject * >( 0 ): opps.front() );
	const double nearest_opp_dist = ( nearest_opp? nearest_opp->distFromSelf(): 1000.0 );
	const Vector2D nearest_opp_pos = ( nearest_opp? nearest_opp->pos(): Vector2D( -1000.0, 0.0 ) );

	Vector2D pass_point;
	if ( Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL ) )
	{
		if ( pass_point.x > wm.self().pos().x - 1.0 )
		{
			bool safety = true;
			const PlayerPtrCont::const_iterator opps_end = opps.end();
			for ( PlayerPtrCont::const_iterator it = opps.begin();
				it != opps_end;
				++it )
			{	
				if ( (*it)->pos().dist( pass_point ) < 4.0 )
				{
				safety = false;
				}
			}

			if ( safety )
			{
				dlog.addText( Logger::TEAM,__FILE__": (execute) do best pass" );
				agent->debugClient().addMessage( "OffKickPass(1)" );
				Body_Pass().execute( agent );
				agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
				return true;
			}
		}
	}

	if ( nearest_opp_dist < 7.0 )
	{
		if ( Body_Pass().execute( agent ) )
		{
			dlog.addText( Logger::TEAM,__FILE__": (execute) do best pass" );
			agent->debugClient().addMessage( "OffKickPass(2)" );
			agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
			return true;
		}
	}

	// dribble to my body dir
	if ( nearest_opp_dist < 5.0&& nearest_opp_dist > ( ServerParam::i().tackleDist()+ ServerParam::i().defaultPlayerSpeedMax() * 1.5 )&& wm.self().body().abs() < 70.0 )
	{
		const Vector2D body_dir_drib_target= wm.self().pos()+ Vector2D::polar2vector(15.0, wm.self().body());

		int max_dir_count = 0;
		wm.dirRangeCount( wm.self().body(), 20.0, &max_dir_count, NULL, NULL );

		if ( body_dir_drib_target.x < ServerParam::i().pitchHalfLength() - 1.0&& body_dir_drib_target.absY() < ServerParam::i().pitchHalfWidth() - 1.0&& max_dir_count < 3)
		{
			// check opponents
			// 10m, +-30 degree
			const Sector2D sector( wm.self().pos(),0.5, 10.0,wm.self().body() - 30.0,wm.self().body() + 30.0 );
				// opponent check with goalie
			if ( ! wm.existOpponentIn( sector, 10, true ) )
			{
				dlog.addText( Logger::TEAM,__FILE__": (execute) dribble to my body dir" );
				agent->debugClient().addMessage( "OffKickDrib(1)" );
				Body_Dribble( body_dir_drib_target,15.0,ServerParam::i().maxDashPower(),10).execute( agent );
				agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
				return true;
			}
		}
	}

	Vector2D drib_target( 50.0, wm.self().pos().absY() );
	if ( drib_target.y < 20.0 ) drib_target.y = 20.0;
	if ( drib_target.y > 29.0 ) drib_target.y = 27.0;
	if ( wm.self().pos().y < 0.0 ) drib_target.y *= -1.0;
	const AngleDeg drib_angle = ( drib_target - wm.self().pos() ).th();

	// opponent is behind of me
	if ( nearest_opp_pos.x < wm.self().pos().x + 1.0 )
	{
		// check opponents
		// 15m, +-30 degree
		const Sector2D sector( wm.self().pos(),0.5, 5.0,drib_angle - 30.0,drib_angle + 30.0 );
		// opponent check with goalie
		if ( ! wm.existOpponentIn( sector, 10, true ) )
		{
		const int max_dash_step= wm.self().playerType().cyclesToReachDistance( wm.self().pos().dist( drib_target ) );
		if ( wm.self().pos().x > 35.0 )
		{
			drib_target.y *= ( 10.0 / drib_target.absY() );
		}

		dlog.addText( Logger::TEAM,__FILE__": (execute) fast dribble to (%.1f, %.1f) max_step=%d",drib_target.x, drib_target.y,max_dash_step );
			agent->debugClient().addMessage( "OffKickDrib(2)" );
			Body_Dribble( drib_target,1.0,ServerParam::i().maxDashPower(),std::min( 5, max_dash_step )).execute( agent );
		}
		else
		{
			dlog.addText( Logger::TEAM,__FILE__": (execute) slow dribble to (%.1f, %.1f)",drib_target.x, drib_target.y );
			agent->debugClient().addMessage( "OffKickDrib(3)" );
			Body_Dribble( drib_target,1.0,ServerParam::i().maxDashPower(),10).execute( agent );
		}
		agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
		return true;
	}

	// opp is far from me
	if ( nearest_opp_dist > 5.0 )
	{
		dlog.addText( Logger::TEAM,__FILE__": opp far. dribble(%.1f, %.1f)",drib_target.x, drib_target.y );
		agent->debugClient().addMessage( "OffKickDrib(4)" );
		Body_Dribble( drib_target,1.0,ServerParam::i().maxDashPower() * 0.4,10).execute( agent );
		agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
		return true;
	}

	// opp is near

	// can pass
	if ( Body_Pass().execute( agent ) )
	{
		dlog.addText( Logger::TEAM,__FILE__": (execute) pass",__LINE__ );
		agent->debugClient().addMessage( "OffKickPass(3)" );
		agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
		return true;
	}

    // opp is far from me
	if ( nearest_opp_dist > 3.0 )
	{
		dlog.addText( Logger::TEAM,__FILE__": (execute) opp far. dribble(%f, %f)",drib_target.x, drib_target.y );
		agent->debugClient().addMessage( "OffKickDrib(5)" );
		Body_Dribble( drib_target,1.0,ServerParam::i().maxDashPower() * 0.2,10).execute( agent );
		agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
		return true;
	}

	if ( nearest_opp_dist > 2.5 )
	{
		dlog.addText( Logger::TEAM,__FILE__": hold" );
		agent->debugClient().addMessage( "OffKickHold" );
		Body_HoldBall().execute( agent );
		agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
		return true;
	}

	{
		dlog.addText( Logger::TEAM,__FILE__": clear" );
		agent->debugClient().addMessage( "OffKickAdvance" );
		Body_AdvanceBall().execute( agent );
		agent->setNeckAction( new Neck_ScanField() );
	}
	return true;

}


bool
Body_BreakThroughPass::get_best_breakthrough_pass(const WorldModel & world,
	Vector2D * target_point,
	double * first_speed,
	int * receiver)
{
	static GameTime S_last_calc_time( 0, 0 );
	static bool S_last_calc_valid = false;
	static Vector2D S_last_calc_target;
	static double S_last_calc_speed = 0.0;
	static int S_last_calc_receiver = Unum_Unknown;

	if ( S_last_calc_time == world.time() )
	{
		if ( S_last_calc_valid )
		{
			if ( target_point )
			{
				*target_point = S_last_calc_target;
			}
			if ( first_speed )
			{
				*first_speed = S_last_calc_speed;
			}
			if ( receiver )
			{
				*receiver = S_last_calc_receiver;
			}
			return true;
		}
		return false;
	}

	S_last_calc_time = world.time();
	S_last_calc_valid = false;

	// create route
	create_routes( world );

	if ( ! S_cached_pass_route.empty() )
	{
		std::vector< PassRoute >::iterator max_it
		= std::max_element( S_cached_pass_route.begin(),
			S_cached_pass_route.end(),
			PassRouteScoreComp() );
		S_last_calc_target = max_it->receive_point_;
		S_last_calc_speed = max_it->first_speed_;
		S_last_calc_receiver = max_it->receiver_->unum();
		S_last_calc_valid = true;
		dlog.addText( Logger::ACTION,
			"%s:%d: get_best_pass() size=%d. target=(%.1f %.1f)"
			" speed=%.3f  receiver=%d"
			,__FILE__, __LINE__,
			S_cached_pass_route.size(),
			S_last_calc_target.x, S_last_calc_target.y,
			S_last_calc_speed,
			S_last_calc_receiver );
	}

	if ( S_last_calc_valid )
	{
		if ( target_point )
		{
			*target_point = S_last_calc_target;
		}
		if ( first_speed )
		{
			*first_speed = S_last_calc_speed;
		}
		if ( receiver )
		{
			*receiver = S_last_calc_receiver;
		}

		dlog.addText( Logger::ACTION,
                      "%s:%d: best pass (%.2f, %.2f). speed=%.2f. receiver=%d"
                      ,__FILE__, __LINE__,
                      S_last_calc_target.x, S_last_calc_target.y,
                      S_last_calc_speed, S_last_calc_receiver );
	}

	return S_last_calc_valid;

}
